URL_FIELD_DESCRIPTION = "Accepts the URL of the public webpage to extract data from."
QUERY_FIELD_DESCRIPTION = "Accepts AgentQL query used to extract the data. The query must be enclosed with curly braces `{}`. Either this field or `prompt` field must be provided."
PROMPT_FIELD_DATA_DESCRIPTION = "Accepts Natural Language description of the data to extract from the page. If AgentQL query is not specified, always use the `prompt` field. Either this field or `query` field must be provided."
PROMPT_FIELD_ELEMENT_DESCRIPTION = "Accepts Natural Language description of the web element to find on the page."
EXTRACT_WEB_DATA_BROWSER_TOOL_DESCRIPTION = "Extracts structured data as JSON from a web page given a URL using either an AgentQL query or a Natural Language description of the data."
EXTRACT_WEB_DATA_TOOL_DESCRIPTION = "Extracts structured data as a JSON from the active web page in a running browser instance using either an AgentQL query or a Natural Language description of the data."
GET_WEB_ELEMENT_BROWSER_TOOL_DESCRIPTION = "Finds a web element on the active web page in a running browser instance using element’s Natural Language description and returns its CSS selector for further interaction, like clicking, filling a form field, etc."



from importlib import metadata

from langchain_agentql.document_loaders import AgentQLLoader
from langchain_agentql.toolkits import AgentQLBrowserToolkit
from langchain_agentql.tools import (
    ExtractWebDataBrowserTool,
    ExtractWebDataTool,
    GetWebElementBrowserTool,
)

try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    # Case where package metadata is not available.
    __version__ = "1.0.0"

__all__ = [
    "AgentQLBrowserToolkit",
    "AgentQLLoader",
    "ExtractWebDataBrowserTool",
    "ExtractWebDataTool",
    "GetWebElementBrowserTool",
    "__version__",
]



""" AgentQL extract web data from browser tool """

from typing import Optional, Type
from typing_extensions import Self

from langchain_community.tools.playwright.base import BaseBrowserTool
from langchain_core.callbacks import (
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)
from pydantic import BaseModel, Field, model_validator

from langchain_agentql.const import (
    DEFAULT_EXTRACT_DATA_TIMEOUT_SECONDS,
    DEFAULT_INCLUDE_HIDDEN_DATA,
    DEFAULT_RESPONSE_MODE,
    DEFAULT_WAIT_FOR_NETWORK_IDLE,
    REQUEST_ORIGIN
)
from langchain_agentql.utils import (
    _aget_current_agentql_page,
    _get_current_agentql_page
)
from langchain_agentql.llm_descriptions import (
    QUERY_FIELD_DESCRIPTION,
    PROMPT_FIELD_DATA_DESCRIPTION,
    EXTRACT_WEB_DATA_BROWSER_TOOL_DESCRIPTION
)
from langchain_agentql.messages import (
    QUERY_PROMPT_REQUIRED_ERROR_MESSAGE,
    QUERY_PROMPT_EXCLUSIVE_ERROR_MESSAGE,
    MISSING_BROWSER_ERROR_MESSAGE
)


class ExtractWebDataBrowserToolInput(BaseModel):
    """Input schema for AgentQL extract web data from browser tool."""

    query: Optional[str] = Field(default=None, description=QUERY_FIELD_DESCRIPTION)
    prompt: Optional[str] = Field(default=None, description=PROMPT_FIELD_DATA_DESCRIPTION)

    @model_validator(mode="after")
    @classmethod
    def check_query_and_prompt(cls, model: Self) -> Self:
        """
        Check that query and prompt cannot be both empty or both provided
        """
        if not model.query and not model.prompt:
            raise ValueError(QUERY_PROMPT_REQUIRED_ERROR_MESSAGE)
        if model.query and model.prompt:
            raise ValueError(QUERY_PROMPT_EXCLUSIVE_ERROR_MESSAGE)
        
        return model


class ExtractWebDataBrowserTool(BaseBrowserTool):
    """AgentQL extract web data from browser tool.

    Setup:
        Install ``langchain-agentql`` and set environment variable ``AGENTQL_API_KEY``.

        .. code-block:: bash

            pip install -U langchain-agentql
            export AGENTQL_API_KEY="your-api-key"

    Instantiation:
        .. code-block:: python

            # sync tool usage
            tool = ExtractWebDataBrowserTool(
                sync_browser=sync_browser,
            )

            # async tool usage
            tool = ExtractWebDataBrowserTool(
                async_browser=async_browser,
            )

    Navigate to the target page with Playwright Navigate Tool:
        .. code-block:: python

            NavigateTool(sync_browser=sync_browser).invoke({"url": "https://www.agentql.com/blog"})

    Invocation with args:
        .. code-block:: python

            tool.invoke({'query': '{ posts[] { title }}'})

        .. code-block:: python

            {'posts': [
                {'title': 'Launch Week Recap—make the web AI-ready'}, 
                {'title': 'Accurate data extraction from PDFs and images with AgentQL'}, 
                {'title': 'Introducing Scheduled Scraping Workflows'}, 
                ...
            ]}
    """  # noqa: E501

    name: str = "extract_web_data_from_browser"
    description: str = EXTRACT_WEB_DATA_BROWSER_TOOL_DESCRIPTION
    args_schema: Type[BaseModel] = ExtractWebDataBrowserToolInput

    timeout: int = Field(default=DEFAULT_EXTRACT_DATA_TIMEOUT_SECONDS)
    """The number of seconds to wait for a request before timing out. Defaults to 900."""
    wait_for_network_idle: bool = Field(default=DEFAULT_WAIT_FOR_NETWORK_IDLE)
    """Whether to wait until the network reaches a full idle state before executing. Defaults to `True`."""
    include_hidden: bool = Field(default=DEFAULT_INCLUDE_HIDDEN_DATA)
    """Whether to take into account visually hidden elements on the page. Defaults to `True`."""
    mode: str = Field(default=DEFAULT_RESPONSE_MODE)
    """'standard' uses deep data analysis, while 'fast' trades some depth of analysis for speed and is adequate for most usecases.
    Learn more about the modes in this guide: https://docs.agentql.com/accuracy/standard-mode. Defaults to 'fast'."""

    def _run(
        self,
        query: Optional[str] = None,
        prompt: Optional[str] = None,
        _: Optional[CallbackManagerForToolRun] = None,
    ) -> dict:
        if not self.sync_browser:
            raise ValueError(MISSING_BROWSER_ERROR_MESSAGE)
        
        page = _get_current_agentql_page(self.sync_browser)
        if query:
            return page.query_data(
                query,
                self.timeout,
                self.wait_for_network_idle,
                self.include_hidden,
                self.mode,
                request_origin=REQUEST_ORIGIN
            )
        elif prompt:
            return page.get_data_by_prompt_experimental(
                prompt,
                self.timeout,
                self.wait_for_network_idle,
                self.include_hidden,
                self.mode,
                request_origin=REQUEST_ORIGIN
            )

    async def _arun(
        self,
        query: Optional[str] = None,
        prompt: Optional[str] = None,
        _: Optional[AsyncCallbackManagerForToolRun] = None,
    ) -> dict:
        if not self.async_browser:
            raise ValueError(MISSING_BROWSER_ERROR_MESSAGE)
        
        page = await _aget_current_agentql_page(self.async_browser)
        if query:
            return await page.query_data(
                query,
                self.timeout,
                self.wait_for_network_idle,
                self.include_hidden,
                self.mode,
                request_origin=REQUEST_ORIGIN
            )
        elif prompt:
            return await page.get_data_by_prompt_experimental(
                prompt,
                self.timeout,
                self.wait_for_network_idle,
                self.include_hidden,
                self.mode,
                request_origin=REQUEST_ORIGIN
            )



from langchain_agentql.tools.extract_web_data_browser_tool import ExtractWebDataBrowserTool
from langchain_agentql.tools.get_web_element_browser_tool import GetWebElementBrowserTool
from langchain_agentql.tools.extract_web_data_tool import ExtractWebDataTool

__all__ = [
    "ExtractWebDataBrowserTool",
    "ExtractWebDataTool",
    "GetWebElementBrowserTool",
]



""" AgentQL extract web data with REST API tool """

import os
from typing import Optional, Type
from typing_extensions import Self

from urllib.parse import urlparse

from langchain_core.tools import BaseTool
from langchain_core.callbacks import (
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)
from pydantic import BaseModel, Field, model_validator

from langchain_agentql.load_data import aload_data, load_data
from langchain_agentql.const import (
    DEFAULT_IS_STEALTH_MODE_ENABLED,
    DEFAULT_WAIT_FOR_PAGE_LOAD_SECONDS,
    DEFAULT_IS_SCROLL_TO_BOTTOM_ENABLED,
    DEFAULT_IS_SCREENSHOT_ENABLED,
    DEFAULT_API_TIMEOUT_SECONDS,
    DEFAULT_RESPONSE_MODE
)
from langchain_agentql.llm_descriptions import (
    EXTRACT_WEB_DATA_TOOL_DESCRIPTION,
    URL_FIELD_DESCRIPTION,
    QUERY_FIELD_DESCRIPTION,
    PROMPT_FIELD_DATA_DESCRIPTION,
)
from langchain_agentql.messages import (
    QUERY_PROMPT_REQUIRED_ERROR_MESSAGE,
    QUERY_PROMPT_EXCLUSIVE_ERROR_MESSAGE,
    UNSET_API_KEY_ERROR_MESSAGE
)


class ExtractWebDataToolInput(BaseModel):
    """Input schema for AgentQL extract web data with REST API tool."""

    url: str = Field(..., description=URL_FIELD_DESCRIPTION)
    query: Optional[str] = Field(default=None, description=QUERY_FIELD_DESCRIPTION)
    prompt: Optional[str] = Field(default=None, description=PROMPT_FIELD_DATA_DESCRIPTION)

    @model_validator(mode="after")
    @classmethod
    def validate_model(cls, model: Self) -> Self:
        """Validate URL scheme and query/prompt requirements."""
        # Validate URL scheme
        url = model.url
        if url:
            parsed_url = urlparse(url)
            if parsed_url.scheme not in ("http", "https"):
                raise ValueError("URL scheme must be 'http' or 'https'")

        # Check that query and prompt cannot be both empty or both provided
        if not model.query and not model.prompt:
            raise ValueError(QUERY_PROMPT_REQUIRED_ERROR_MESSAGE)
        if model.query and model.prompt:
            raise ValueError(QUERY_PROMPT_EXCLUSIVE_ERROR_MESSAGE)

        return model
    

class ExtractWebDataTool(BaseTool):
    """AgentQL extract web data with REST API tool.

    Setup:
        Install ``langchain-agentql`` and set environment variable ``AGENTQL_API_KEY``.

        .. code-block:: bash

            pip install -U langchain-agentql
            export AGENTQL_API_KEY="your-api-key"

    Instantiation:
        .. code-block:: python

            tool = ExtractWebDataTool()

    Invocation with args:
        .. code-block:: python

            tool.invoke({'url': 'https://www.agentql.com/blog', 'query': '{ posts[] { title }}'})

        .. code-block:: python

            {'data': 
                    {'posts': [
                        {'title': 'Launch Week Recap—make the web AI-ready'}, 
                        {'title': 'Accurate data extraction from PDFs and images with AgentQL'}, 
                        {'title': 'Introducing Scheduled Scraping Workflows'}, 
                        ...
                    ]}, 
                'metadata': {'request_id': 'xxxxxx-xxxx-xxxx-xxxx-xxxx'}
            }
    """  # noqa: E501

    name: str = "extract_web_data_with_rest_api"
    description: str = EXTRACT_WEB_DATA_TOOL_DESCRIPTION
    args_schema: Type[BaseModel] = ExtractWebDataToolInput

    api_key: Optional[str] = Field(default=None)
    """AgentQL API key. You can create one at https://dev.agentql.com."""
    timeout: int = Field(default=DEFAULT_API_TIMEOUT_SECONDS)
    """The number of seconds to wait for a request before timing out. Defaults to 900."""
    is_stealth_mode_enabled: bool = Field(default=DEFAULT_IS_STEALTH_MODE_ENABLED)
    """Whether to enable experimental anti-bot evasion strategies. This feature may not work for all websites at all times. 
    Data extraction may take longer to complete with this mode enabled. Defaults to `False`."""
    wait_for: int = Field(default=DEFAULT_WAIT_FOR_PAGE_LOAD_SECONDS)
    """The number of seconds to wait for the page to load before extracting data. Defaults to 0."""
    is_scroll_to_bottom_enabled: bool = Field(default=DEFAULT_IS_SCROLL_TO_BOTTOM_ENABLED)
    """Whether to scroll to bottom of the page before extracting data. Defaults to `False`"""
    mode: str = Field(default=DEFAULT_RESPONSE_MODE)
    """'standard' uses deep data analysis, while 'fast' trades some depth of analysis for speed and is adequate for most usecases.
    Learn more about the modes in this guide: https://docs.agentql.com/accuracy/standard-mode. Defaults to 'fast'."""
    is_screenshot_enabled: bool = Field(default=DEFAULT_IS_SCREENSHOT_ENABLED)
    """Whether to take a screenshot before extracting data. Returned in 'metadata' as a Base64 string. Defaults to `False`"""
    
    _params: dict = {}
    _metadata: dict = {}
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._params = {
            "wait_for": self.wait_for,
            "is_scroll_to_bottom_enabled": self.is_scroll_to_bottom_enabled,
            "mode": self.mode,
            "is_screenshot_enabled": self.is_screenshot_enabled,
        }
        self._metadata = {
            "experimental_stealth_mode_enabled": self.is_stealth_mode_enabled,
        }
        self._api_key = self.api_key or os.getenv("AGENTQL_API_KEY")
        if not self._api_key:
            raise ValueError(UNSET_API_KEY_ERROR_MESSAGE)
        
    def _run(
        self,
        url: str,
        query: Optional[str] = None,
        prompt: Optional[str] = None,
        _: Optional[CallbackManagerForToolRun] = None,
    ) -> dict:
        return load_data(
            url=url,
            query=query,
            prompt=prompt,
            params=self._params,
            api_key=self._api_key,
            timeout=self.timeout,
            metadata=self._metadata,
        )

    async def _arun(
        self,
        url: str,
        query: Optional[str] = None,
        prompt: Optional[str] = None,
        _: Optional[AsyncCallbackManagerForToolRun] = None,
    ) -> dict:
        return await aload_data(
            url=url,
            query=query,
            prompt=prompt,
            params=self._params,
            api_key=self._api_key,
            timeout=self.timeout,
            metadata=self._metadata,
        )
